#!perl
# PureMessage OS Health Reports
# Base OS Module
# Never initialize this module directly; use an operating system
# sub-module instead.

#
# Configuration attributes for ::OS objects:
#
# delay: (default 60) - how often to collect statistics from children,
#   and how often reentrant children should run
#
# step: (default 1) - how often continuous pipelines should be instructed
#   to poll their subprocesses for statistics
#


package PureMessage::Reports::OS;

use 5.006;
use strict;
use warnings;
use Exporter;
use base qw(Exporter);

use PureMessage::Reports::OS::Pipeline::ReEntrant;

our @EXPORT_OK = qw(%uname_to_module);

our $VERSION = '5.500';

our %uname_to_module = (
    "linux"     =>      "Linux",
    "solaris"   =>      "SunOS",
    "freebsd"   =>      "FreeBSD",
    "aix"       =>      "AIX",
    "hpux"      =>      "HPUX",
);

return 1;

sub new_by_os {
    my($class, %args) = (@_);
    
    unless($uname_to_module{$^O}) {
        warn "no OS Health module available for \"$^O\"!";
        return;
    }
    
    $class = join("::", "PureMessage::Reports::OS", $uname_to_module{$^O});
    
    unless(eval("use $class; 1;")) {
        warn "loading of $class failed: $@";
        return;
    }
    
    return $class->new(%args);
}

sub new {
    my($class, %args) = (@_);
    
    %args = (
        cycle	=> 0,
        start	=> time(),
        delay	=> 60,
        step	=> 1,
    
        %args,
    );
    
    my $self = bless \%args, $class;
    $self->{sample_average} ||= $self->{delay} / $self->{step}; # /

    return $self;
}

# TODO: error checking etc in here

sub run {
    my $self = shift;
    
    foreach my $pipe (@{$self->{pipelines}})
    {
        $pipe->check; $pipe->check;
    }
    
    if(!grep(!$self->{$_}, @{$self->{tags}})) {
        $self->{cycle}++;
        $self->{'time'} = time();
        map { $self->{$_} = 0; } (@{$self->{tags}});
        return $self->stats;
    } else {
        return;
    }
}

sub start {
    my $self = shift;
    
    foreach my $pipe (@{$self->{pipelines}}) {
        $pipe->start;
    }
    
    $self->{started} = 1;
    
    return;
}

sub started {
    $_[0]->{started} || 0;
}

sub stop {
    my $self = shift;
    
    foreach my $pipe (@{$self->{pipelines}}) {
        $pipe->stop;
    }
    
    delete($self->{started});
    
    return;
}

sub stats {
    my $self = shift;
    return {
        map
            { ($_ => $self->{$_}) }
            (qw(
                cpu_free ram_free swap_free uptime load_avg
                disc_used disc_free disc_read disc_write
                cycle time start
            ))
    };
}

sub calc_uptime {
    my($self, $str) = @_;
    
    my $up = 0;
    my(@segments) = split(/\s*,\s*/, $str);
    
    foreach my $s (@segments) {
     if($s =~ m{^([0-9]+)\+([0-9]{2}):([0-9]{2}):([0-9]{2})}) {
         $up = ($1 * 86400) + ($2 * 3600) + ($3 * 60) + $4;
     } elsif($s =~ m{^([0-9]{2}):([0-9]{2}):([0-9]{2})}) {
         $up = ($1 * 3600) + ($2 * 60) + $3;
     } elsif($s =~ m{^([0-9]+) days$}) {
         $up += $1 * 86400;
     } elsif($s =~ m{^([0-9]+):([0-9]+)$}) {
         $up += ($1 * 3600) + ($2 * 60);
     } elsif($s =~ m{^([0-9]+) min$}) {
         $up += $1 * 60;
     }
    }
    
    return $up;
}

# this, thank god, works the same on all OSes in question.
sub parse_uptime {
    my($self, $pipe) = @_;
    my $data = $pipe->data;
    
    if($data =~ m{up (.+?),\s*[0-9]+\s*users?,\s+load average: ([\.0-9]+),}) {
        $self->{load_avg} = $2;
        $self->{uptime} = $self->calc_uptime($1);
        $self->{parse_uptime}++;
    }
}

sub pipeline_uptime {
    my($self, %args) = (@_);
    
    return PureMessage::Reports::OS::Pipeline::ReEntrant->new(
        program => [ "/usr/bin/uptime" ],
        callback_exit => sub { $self->parse_uptime(@_); },
        delay => $self->{delay},
        %args
    );
}

# "df -k"'s output looks pretty much the same on all the UNIXes
# we're concerned with except HP-UX.
sub parse_df {
    my($self, $pipe) = @_;
    
    $self->{disc_used} = 0;
    $self->{disc_free} = 0;
    
    my $data = $pipe->data;
    my(@data) = split(/\n/s, $data);
    shift(@data);
    
    while(my $line = shift(@data)) {
        $self->{parse_df}++;
        
        my(@line) = split(/\s+/, $line);

        if(@line == 1) {
            # if the path's too long, the line gets split
            my $lineb = shift(@data);
            push(@line, split(/\s+/, $lineb));
        }
        
        # local discs only, no procfs or nfs
        if(substr($line[0], 0, 1) eq '/') {
            # on AIX output is slightly different
            if(substr($line[3], -1) eq '%') {
                $self->{disc_free} += $line[2];
                $self->{disc_used} += $line[1] - $line[2];
            } else {
                $self->{disc_used} += $line[2];
                $self->{disc_free} += $line[3];
            }
        }
    }
}

sub pipeline_df {
    my $self = shift;
    return PureMessage::Reports::OS::Pipeline::ReEntrant->new(
        program => [ "/bin/df", "-k" ],
        callback_exit => sub { $self->parse_df(@_); },
        delay => $self->{delay},
    );
}

sub stat_ready {
    my($self, $stat) = @_;
    if($self->{"${stat}_rows_in"} &&
        ($self->{"${stat}_rows_in"} % $self->{sample_average} == 0)) {
        $self->{"parse_${stat}"}++;
        return 1;
    } else {
        return;
    }
}

=pod

=head1 NAME

PureMessage::Reports::OS - Base class for OS Health Monitors

=head1 SYNOPSIS

  my $osh = PureMessage::Reports::OS->new_by_os(step => 5, delay => 30);

  $osh->start;

  while(1) {
          sleep(1);
          if(my $stats = $osh->run) {
                  [do something with statistics]
          }
  }

=head1 DESCRIPTION

This is the base class for OS Health monitors. Every operating system reports
it's hardware usage slightly differently and it is the job this class's children
to gather this information and present it in a uniform format.

Information is gathered by running system software and analyzing the output.
Some of this software provides continuous output, returning a line (or several
lines) of information. Other software must be re-run each time the current
statistics are desired.

=head1 CONSTRUCTOR

The new_by_os() class method determines what operating system we are running
on, loads the correct class, and initalises it with new(). The following
arguments are useful under all operating systems as parameters to new_by_os
or new:

=over

=item delay

How often to return statistics from underlying software, in seconds. This
value also determines how often software which must be re-run (such as "df")
is executed.

=item step

How often to instruct continuous-feed software to provide us with statistics.
These statistics are averaged before they are returned.

=back

=head1 METHODS

=over

=item start

Start up all of the processes that will be used for os health data collection
on this operating system.

=item stop

Shut down any processes currently being used for os health data collection.

=item run

Check to see if there are any new statistics to be collected. If the delay
has elapsed, return the current set of statistics as a hashref.

=item stats

Return the current set of statistics as a hashref. Not all statistics
are available on all operating systems. The following key/value pairs
will be returned:

=over

=item cpu_free

How much CPU this system has available, in percent.

=item ram_free

How much RAM this system has available, in kilobytes.

=item swap_free

How much swap this system has available, in kilobytes.

=item uptime

How long this system has been up for, in seconds.

=item load_avg

The current load average of the system. (How many cycles worth of work
the system wants to do per actual cpu cycle.)

=item disc_used

Total hard disc space used across all local partitions, in kilobytes.

=item disc_free

Total hard disc space free across all local partitions, in kilobytes.

=item disc_read

Total amount read from the hard disc, in kilobytes. 

=item disc_write

Total amount written to the hard disc, in kilobytes.

Not all operating systems supply seperate disc read and write statistics;
on those systems, disc_read and disc_write will each be set to half of the
reported throughput.

=item cycle

How many times we have collected statistics so far. The first time run()
returns statistics, cycle will be set to 1, the second time, it will be
incremented to 2, etc.

=item time

UNIX timestamp of when the last statistics-gathering cycle completed.

=item start

UNIX timestamp of when this object was initalised.

=back

=back

=head1 SEE ALSO

L<PureMessage::Reports::OS::Reporter>, 
L<PureMessage::Reports::OS::Pipeline::Continuous>,
L<PureMessage::Reports::OS::Pipeline::ReEntrant>

=head1 NOTES

Some software supplies more than one useful statistic. For example, on Solaris,
"vmstat" supplies both memory and CPU usage. Under Linux, the "vmstat" program
supplies disc IO information as well.

"uptime" is run and parsed exactly the same on every system. Invocation
of "df" varies from OS-to-os, but the output is parsed in exactly the
same way on each OS, with the exception of AIX

Continuous feed software is preferred over software that must be re-run.
The "PureMessage::Reports::OS::Pipeline::Continuous" class manages continuous
feed software, and the "PureMessage::Reports::OS::Pipeline::ReEntrant" classes
manages software that must be re-run.

The quality of available statistics differs between operating systems. In order
to ensure usability on as many systems as possible, only software that is
shipped with the operating systems (and does not require any daemons to be
running) is used by these classes.

=over

=item Linux

provides the most complete set of information available.

=item Solaris

There is no stock Solaris software that will report how much RAM is on a
system; only how much you have left. Because of this, PureMessage::Reports::OS
works the same way.

=item FreeBSD

does not provide seperate disc throughput statistics for read and write.
PureMessage::Reports::OS will take the total throughput value and report
"disc_read" and "disc_write" to each be half of that.

FreeBSD also does not aggregate it's disc throughput statistics; you get one
statistic for each disc and have to add them up yourself. On systems with
many discs installed, the activity of some of them may not be reported.

=item HPUX

does not provide seperate disc throughput statistics for read and write either.
It also does not provide seperate swap and ram usage information; "swap_free"
will always be returned as "0" on HPUX.

=item AIX

does not provide seperate swap and ram usage information; "swap_free" will
always be returned as "0" on AIX.

=back

=cut
