#!perl

package PureMessage::Reports::OS::FreeBSD;

use strict;
use warnings;
use PureMessage::Reports::OS;
use base qw(PureMessage::Reports::OS);

use PureMessage::Reports::OS::Pipeline::ReEntrant;
use PureMessage::Reports::OS::Pipeline::Continuous;

our(@tags) = qw(parse_df parse_vmstat parse_iostat parse_top);

return 1;

sub new {
    my($class, %args) = (@_);
    my $self = $class->SUPER::new(disc_read_x => 0, disc_write_x => 0, %args);

    bless $self, $class;
    
    if(!$self->{pipelines}) {
        $self->{pipelines} = [
            $self->pipeline_df,
        
            PureMessage::Reports::OS::Pipeline::Continuous->new(
                program => [ "/usr/bin/vmstat", "-n", "0", $self->{step} ],
                callback => sub { $self->parse_vmstat(@_); },
                # first line is "summary" since boot which we want to avoid
                preamble => qr{[0-9]\n},
                eod => qr{\n},
            ),
            
            PureMessage::Reports::OS::Pipeline::ReEntrant->new(
                program => [ "/usr/bin/top", "-b", "0" ],
                callback_exit => sub { $self->parse_top(@_); },
                delay => $self->{delay},
            ),
                
            PureMessage::Reports::OS::Pipeline::Continuous->new(
                program => [
                                "/usr/sbin/iostat", "-I", "-t", "da",
                                $self->{step}
                           ],
                callback => sub { $self->parse_iostat(@_); },
                eod => qr{\n},
            ),
        ];
    }

    $self->{tags} ||= \@tags;
    
    $self;
}

sub calc_kscale {
    my($self, $val) = @_;
    my $valk = '';
    
    if($val =~ s{([A-Z])$}{}) {
        my $vals = uc($1);

        if($vals eq 'M') {
            $valk = $val * 1024;
        } elsif($vals eq 'G') {
            $valk = $val * 1048576;
        } else {
            $valk = $val;
        }
    } else {
        $valk = $val;
    }
    
    return $valk;
}

sub parse_top {
    my($self, $pipe) = @_;
    my $data = $pipe->data;
    
    my(@data) = split(/\n/, $data);
    
    if($data[0] =~ m{load averages:\s*([\.0-9]+),.+up ([+:0-9]+)\s}) {
        $self->{parse_top}++;
        $self->{load_avg} = $1;
        $self->{uptime} = $self->calc_uptime($2);
    }        
    
    if($data[3] =~ m{^Mem.*\s([A-Za-z0-9]+) Free}) {
        $self->{parse_top}++;
        $self->{ram_free} = $self->calc_kscale($1);
    }
    
    if($data[4] =~ m{^Swap.*\s([A-Za-z0-9]+) Free}) {
        $self->{parse_top}++;
        $self->{swap_free} = $self->calc_kscale($1);
    }
}

sub parse_vmstat {
    my($self, $pipe, $data) = @_;
    
    $data =~ s{^\s+}{};

    return if($data =~ m{^[a-z]});

    my(@data) = (split(/\s+/, $data));

    if(defined($data[16])) {
        $self->{cpu_idle_x} += $data[16];
        $self->{vmstat_rows_in} ++;

        if($self->stat_ready("vmstat")) {
            $self->{cpu_free} = $self->{cpu_idle_x} / $self->{sample_average}; # /
            $self->{cpu_idle_x} = 0;
        }
    } else {
        warn "malformed vmstat line \"$data\"";
    }
    
}

sub parse_iostat {
    my($self, $pipe, $data) = @_;
    
    $data =~ s{^\s+}{};
    
    return if(substr($data, 0, 3) eq "tty");
    
    my(@data) = split(/\s+/, $data);
    
    if($data[0] eq "tin") {
        $self->{mb_pos} = [ grep($data[$_] eq 'MB', $[ .. $#data) ];
        return;
    }
    
    if($data[0] !~ m{^[0-9]}) {
        warn "malformed iostat line \"$data\"";
        return;
    }
    
    if(!$self->{_warned_mb_pos} && (!$self->{mb_pos})) {
        $self->{_warned_mb_pos} = 1;
        warn "got iostat statistics before we got any headers! \"$data\"";
        return;
    } elsif(!@{$self->{mb_pos}}) {
        return;
    }
    
    my $kb = 0;
    map { $kb += $data[$_] } (@{$self->{mb_pos}});
    $kb = $kb * 1024;
    
    # freeBSD doesn't give us seperate read/write stats... :/
    $self->{disc_read_x} += $kb/2;
    $self->{disc_write_x} += $kb/2;
    $self->{iostat_rows_in}++;
    
    if($self->stat_ready("iostat")) {
        $self->{disc_read} = $self->{disc_read_x};
        $self->{disc_write} = $self->{disc_write_x};
        $self->{disc_read_x} = 0;
        $self->{disc_write_x} = 0;
    }
}
