#!perl

package PureMessage::Reports::OS::Reporter;

use strict;
use warnings;
use vars qw(%ENV);
use DBI;
use PureMessage::Reports::Util qw(get_hostname);
use PureMessage::Reports::OS;
use PureMessage::Reports::DB;

use constant AOP        => 0;
use constant ALABEL     => 1;
use constant ATYPE      => 2;
use constant ASUB       => 3;

return 1;

# change CPU idle % into CPU utilized %, multiply by 10 to preserve decimal
sub munge_cpu_free { (100 - $_[0]) * 10; }
# multiply by 100 to keep 2 decimal places
sub munge_load_avg { $_[0] * 100; }

# methods

sub new {
    my($class, %args) = @_;
    
    if(ref($args{data})) {
        my $data = delete($args{data});
        %args = (%args, %{$data});
    }
    
    %args = (
        duration => 300,
    
        aggregate => {
            disc_kbytes_written         => [ "+", "disc_write", "absolute" ],
            disc_kbytes_read            => [ "+", "disc_read", "absolute" ],
            disc_kbytes_allocated       => [ "/", "disc_used", "absolute" ],
            ram_allocated               => [ "/", "ram_free", "absolute" ],
            swap_allocated              => [ "/", "swap_free", "absolute" ],
            cpu_utilized                => [ "/", "cpu_free", "relative", \&munge_cpu_free ],
            load                        => [ "/", "load_avg", "relative", \&munge_load_avg ],
            uptime                      => [ "l", "uptime", "absolute" ]
        },
        
        osconfig => {}, 

        %args,
    );

    my $self = bless \%args, $class;

    $self->{_started} = 0;
    $self->{dbh} ||= get_dbh();
    $self->{label_index} ||= $self->stats_label_index;
    $self->{start_t} ||= time() + $self->time_until_next_window;
    $self->{osh} ||= PureMessage::Reports::OS->new_by_os(%{$args{osconfig}});
    $self->{hostname} ||= get_hostname($self->{'dbh'});

    return unless($self->{dbh} && $self->{label_index} && $self->{osh});
    
    $self;
}

sub run {
    my $self = shift;
    my $poll_status = 0;
    my $error_msg = "";

    unless ($self->{_started}) {
        my $start_status = $self->start();
        defined $start_status or $error_msg = "Error in start()";
    }

    if($self->{_started} and $self->{osh}->started()) {
        $poll_status = $self->poll(@_);
        defined $poll_status or $error_msg = "Error in poll()";
    }

    if ($error_msg) {
        warn $error_msg; # will effectively log it
        die $error_msg;
    }

    return $poll_status;
}

sub shutdown {
    my $self = shift;
    $self->{osh}->stop;
    delete($self->{osh});
    delete($self->{dbh});
    bless $self, 'main::perlWontLetYouUnbless';
    return;
}

# returns undefined only on error
sub poll {
    my $self = shift;
    if(my $stats = $self->{osh}->run()) {
        $self->update_stats($stats);
    }
    
    if(time() > $self->{end_t}) {
        if(my($id, $start_t, $end_t) = ($self->find_step)) {
            if($self->{id} != $id) {
                my $oid = $self->{id};
                $self->{id} = $id;
                $self->{start_t} = $start_t;
                $self->{end_t} = $end_t;

                if(my(%stats) = ($self->aggregate_stats)) {
                    return $self->insert_stats($oid, %stats);
                } else {
                    warn "End of period, but there are no stats to aggregate!";
                    return;
                }

            } else {
                warn "Strange: we are past our reporting period, ", 
                    "but reporting ID has not changed.";
                return 0;
            }
        } else {
            warn "Failed to get next step.";
            return;
        }
    } else {
        return 0;
    }
}

sub time_until_next_window {
    my $self = shift;
    return $self->{duration} - (time() % $self->{duration});
}

# returns undefined only on error
sub start {
    my $self = shift;
    return 0 if $self->{_started};

    # don't start unless we're close to a {duration}-second window boundary
    # (close = 2 seconds)
    my $delta_t = $self->time_until_next_window;
    return 0 unless ($delta_t <= 2 || $delta_t >= ($self->{duration} - 1));

    my($id, $start_t, $end_t) = ($self->find_next_step);
    return unless($start_t);

    $self->{id} = $id;
    $self->{start_t} = $start_t;
    $self->{end_t} = $end_t;
    $self->{osh}->start();
    $self->{_started} = 1;

    return 1;
}

sub stats_label_index {
    my $self = shift;
    my %rv;
    
    if(my $sth = $self->{dbh}->prepare_cached(q{
        SELECT name, id FROM prd_runtime_fact_type
    })) {
        if($sth->execute()) {
            while(my($k, $v) = ($sth->fetchrow_array)) {
                $rv{$k} = $v;
            }

            $sth->finish();
            return \%rv;
        } else {
            warn "Failed to load runtime fact type table: ", DBI::errstr;
            return;
        }
    } else {
        warn "Failed to prepare to load runtime fact type table: ", DBI::errstr;
        return;
    }
}

sub update_stats {
    my $self = shift;
    $self->{stats} ||= [];
    push(@{$self->{stats}}, @_);
}

sub aggregate_stats {
    my $self = shift;
    my(%stats);
    
    my $nstats = scalar(@{$self->{stats}});
    return unless $nstats;
    
    keys %{$self->{aggregate}};		# reset iterator
    foreach my $stat (@{$self->{stats}}) {
        while(my($k, $v) = each(%{$self->{aggregate}})) {
            $stats{$k} += $stat->{$v->[ALABEL]};
        }
    }
    
    while(my($k, $v) = each(%{$self->{aggregate}})) {
        if($v->[AOP] eq "/") {
            $stats{$k} = $stats{$k} / $nstats;
        } elsif($v->[AOP] eq "l") {
            $stats{$k} = $self->{stats}->[$#{$self->{stats}}]->{$v->[ALABEL]};
        }

        if($v->[ASUB]) {
            $stats{$k} = $v->[ASUB]->($stats{$k});
        }
    }
    
    $self->{stats} = [];
    $self->{last_aggregate} = \%stats;
    
    return(%stats);
}

sub insert_stats {
    my($self, $id, %stats) = (@_);
    
    if(!%stats) {
        warn "No stats to insert!";
        return;
    }
    
    my $inserted = 0;
    
    $self->{dbh}->begin_work if($self->{dbh}->{AutoCommit});
    while(my($k, $v) = each(%stats)) {
        $v = int($v);
        if(my $atype = $self->{aggregate}->{$k}->[ATYPE]) {
            if(my $sid = $self->{label_index}->{$k}) {
                my $query = qq|
                    INSERT INTO prd_pmx_runtime_fact
                           ( period_id,
                             runtime_fact_type_id,
                             pmx_host_id,
                             value_${atype}
                           )
                           SELECT $self->{id},
                                  $sid,
                                  prd_pmx_host.id,
                                  $v
                           FROM   prd_pmx_host
                           WHERE  name = '$self->{'hostname'}';
                |;
                
                if(my $sth = $self->{dbh}->do($query)) {
                    # success!!!
                    $inserted++;
                } else {
                    warn "Preparing to insert OS health into db failed:",
                        DBI::errstr;
                    $self->{dbh}->rollback;
                    return;
                }
            } else {
                warn "Can't find a label ID for statistic \"$k\"!";
                $self->{dbh}->rollback;
                return;
            }
        } else {
            warn "No data type found for statistic \"$k\"!";
            $self->{dbh}->rollback;
            return;
        }
    }
    
    $self->{dbh}->commit;
    return $inserted;
}

sub find_step {
    my($self, $timestamp) = @_;
    $timestamp ||= time();
    my $utimestamp = $timestamp - ($timestamp % ($self->{duration}));
    
    if(my $sth = $self->{dbh}->prepare_cached(q{
        SELECT id,
           extract(epoch from period_start) AS period_start_t,
           extract(epoch from period_end) AS period_end_t
           FROM prd_period
           WHERE duration = ?
           AND period_start = ?
    })) {
        if($sth->execute(
               $self->{duration},
               scalar(gmtime($utimestamp)) . " GMT"
        )) {
            if(my($id, $start_t, $end_t) = ($sth->fetchrow_array)) {
                $sth->finish();
                return($id, $start_t, $end_t);
            } else {
                warn "Could not find a period that matches duration ",
                    $self->{duration}, " for time ",
                    scalar(localtime($timestamp)), " (starting ",
                    scalar(gmtime($utimestamp)), " GMT)!";
                return;
            }
        } else {
            warn "Could not look up reporting period in database: ",
                DBI::errstr;
            return;
        }
    } else {
        warn "Could not obtain a handle to look up reporting period in database: ",
            DBI::errstr;
        return;
    }
}

sub find_next_step {
    my $self = shift;
    return($self->find_step(time() + $self->duration));
}

sub last_aggregate {
    shift->{last_aggregate};
}

sub duration {
    shift->{duration};
}

sub osh {
	shift->{osh};
}

### private functions

=pod

=head1 NAME

PureMessage::Reports::OS::Reporter - Gather and store OS Health Information

=head1 SYNOPSIS

  my $reporter = PureMessage::Reports::OS::Reporter->new();

  while(1) {
          $reporter->run;
          sleep(1);
  }

  $reporter->shutdown;

  exit(0);

=head1 DESCRIPTION

The PureMessage::Reports::OS::Reporter class polls a
PureMessage::Reports::OS object, and inserts the information returned from
this object into an SQL database at regular intervals.

=head1 CONSTRUCTION

PureMessage::Reports::OS::Reporter objects are constructed with new();
the following parameters are useful:

=over

=item duration

(default: 300) How often information should be written into the SQL database.
Time intervals with the same duration value must already exist in the database;
see pmx-reports-set-time-ranges.

=item osconfig

(default: empty) A HASH reference containing any extra parameters that
should be passed to the L<PureMessage::Reports::OS> OS-specific health
collection module.

=back

There are also other parameters which may be specified that are not normally
useful, but can be used to affect the reporter's behaviour:

=over

=item dbh

A DBI database connection handle to use instead of the default one supplied
by PureMessage::Reports::DB

=item osh

A PureMessage::Reports::OS object to use instead of creating a new one.

=back

=head1 METHODS

=head2 Main Methods:

=over

=item run()

Start up the OS health monitor if it is time to start it. Poll the OS health
monitor for information if it is already started. This method should be run
as often as possible, and at least 10 times for each cycle's duration.

=item shutdown()

Closes the database connection, shuts down the OS health object, and
invalidates the Reporter object.

=back

=head2 Other Useful Methods:

=over

=item start()

Wait until the beginning of the next time range, and start up the OS health
monitor. Use with caution; it could take up to as long as the object's
duration for this operation to complete.

=item poll()

Poll the OS health monitor for information, and insert aggregated statistics
into the database if it is time to do so.

=item find_step([timestamp])

Finds the time range in the database that the supplied timestamp falls within.
If no timestamp is specified, the current time is used. Returns an array
containing the database ID #, start time, and end time of the time range found,
or undef if no time range could be found.

=item find_next_step

Finds the next time range in the database. Equivalent to
$reporter->find_step(time() + $reporter->duration);

=item osh

Returns the underlying PureMessage::Reports::OS object.

=back

=head1 SEE ALSO

L<PureMessage::Reports::OS>, L<pmx-reports-set-time-ranges>,
L<PureMessage::Reports::DB>

=cut
